
/***********************************************/
/* Run Text Analytics Category Model in Hadoop */
/***********************************************/

options set=SAS_HADOOP_JAR_PATH="/opt/sasinside/hadoop/lib";
options set=SAS_HADOOP_CONFIG_PATH="/opt/sasinside/hadoop/conf";

libname dblib hadoop server = "nodename.demo.xyz.com" user=your.name password=xyz123;

* INPUT DATASET;
%let input_ds = dblib.yourdata;
%let document_column = text;
    
* OUTPUT DATASET;
%let tmp_output_position_ds = dblib.out_ds_cat;
%let output_position_ds = out_ds_cat;
%let output_document_ds = out_docs_ds;

* DELETE OUTPUT BEFORE START, (OPTIONAL);
proc delete data=&tmp_output_position_ds; run;

proc ds2 ds2accel=yes;
    require package tkcat; run;
    require package tktxtanio; run;

    thread workerth / overwrite=YES;

        dcl package tkcat cat();
        dcl package tktxtanio txtanio();

        dcl double docid;
        dcl double _start_offset;
        dcl double _end_offset;
        dcl char(1024) _name;
        dcl char(1024) _full_path;
        dcl char(1024) _term;
        dcl binary(8) _apply_settings;
        dcl binary(8) _mco_binary;
        dcl binary(8) _trans;
        dcl binary(8) _document;
        dcl double _status;
        dcl double _num_matches;
        dcl double _i;
        dcl double _num_terms;
        dcl double _j;

        retain _apply_settings;
        retain _mco_binary;
        retain _trans;

        /* only runs once when starting */
        method init();
            _apply_settings = cat.new_apply_settings();

            _mco_binary = txtanio.new_local_file('/user_space/categories.mco');
            
            cat.set_categories_model(_apply_settings, _mco_binary);
            cat.set_return_match_positions_for_categories(_apply_settings, 1);
            cat.set_relevancy_type(_apply_settings, 1);

            _status = cat.initialize_categories(_apply_settings);
            if _status NE 0 then put 'ERROR: initialize_categories fails';

            _trans = cat.new_transaction();
        end;

        /* run per row of input */
        method run();
            set &input_ds(keep=(&document_column docid));

            _document = txtanio.new_document_from_string(&document_column);

            /* set the document on the transaction so we're ready to process */
            _status = cat.set_document(_trans, _document);
            if _status NE 0 then put 'ERROR: set_document fails on document:' docid;

            /* apply the binary to the document */
            _status = cat.apply_categories(_apply_settings, _trans);
            if _status NE 0 then put 'ERROR: apply_categories fails on document:' docid 'the observation variable is probably empty.';

            /* get the number of categories matched */
            _num_matches = cat.get_nb_matched_categories(_trans);
            _i = 0;
            do while (_i LT _num_matches);
                /* we used the name for both name & full_path for better alignment with LITI code */
                _name = cat.get_category_name(_trans, _i);
                _full_path = _name;

                _num_terms = cat.get_nb_matched_terms_category(_trans, _i);

                _j = 0;
                do while (_j LT _num_terms);
                    _start_offset = cat.get_matched_term_start_offset_category(_trans, _i, _j);
                    _end_offset = cat.get_matched_term_end_offset_category(_trans, _i, _j);
                    _term = cat.get_matched_term_category(_trans, _i, _j);

                    output;

                    _j = _j + 1;
                end;

                _i = _i + 1;
            end;

            /* clean up resources */
            cat.clean_transaction(_trans);
            txtanio.free_object(_document);
        end;

        /* only runs once at end */
        method term();
            /* clean up resources */
            cat.free_transaction(_trans);
            cat.free_apply_settings(_apply_settings);
            txtanio.free_object(_mco_binary);
        end;
    endthread;
run;

/* now create output data set */
data &tmp_output_position_ds(
         keep=(
           docid
           _name
           _full_path
           _start_offset
           _end_offset
           _term
         )
         overwrite=yes
       );
dcl THREAD workerth THRD;
method run();
set from THRD;
end;
enddata;
run;
quit;


data &output_position_ds;
 set &tmp_output_position_ds(
   rename=(
            docid=document_id
            _name=name
            _full_path=full_path
            _start_offset=start_offset
            _end_offset=end_offset
            _term=term
          ));
run;


proc datasets lib=work memtype=data;
   modify &output_position_ds;
     attrib _all_ label=' ';
     attrib _all_ format=;
contents data=&output_position_ds;
run;
quit;

data &output_position_ds;
  retain document_id name full_path start_offset end_offset term;
  set &output_position_ds;
run;

/* ------------- */
/* ------------- */
/* ------------- */

/* get unique category name and document ID pairs */

proc sort data=&output_position_ds out=_tmp_doc_cat_ds(rename=(document_id = _document_)) nodupkey;
    by name document_id;
run;

/* assign a column name to each category */

data _tmp_doc_cat_ds (keep = name _document_ column_name in_category_rule);
    length column_name $8.;

    set _tmp_doc_cat_ds;
    by name _document_;

    retain previous_name column_name index;

    if _n_ = 1 then do;
        previous_name = name;
        column_name = "c_1";
        index = 1;
    end;

    if name NE previous_name then do;
        index = index + 1;
        previous_name = name;
        column_name = "c_" || kleft(index);
    end;

    in_category_rule = 1;
run;

/* sort on document id */
 
proc sort data=_tmp_doc_cat_ds;
    by _document_ name;
run;

/* transpose so that we have a column for each category */

proc transpose data = _tmp_doc_cat_ds out = _tmp_transposed_ds(drop=_NAME_);
    var in_category_rule;
    by _document_;
    id column_name;
    idlabel name;
run;

/* merge the category columns with the input data set */
 
data _tmp_missing_ds;
    merge _tmp_transposed_ds &input_ds(rename=(docid=_document_));
    by _document_;
run;

/* set missing values to zero */

data &output_document_ds;
    set _tmp_missing_ds;
    array ss[*] c_ : ;
    do j = 1 to dim(ss);
    if ss(j) = . then ss(j) = 0;
    end;
    drop j;
run;

/* sort the position data set */

proc sort data=&output_position_ds;
    by document_id name;
run;

/* map the column name to the category name in the position data set */

data &output_position_ds;
    merge &output_position_ds _tmp_doc_cat_ds(rename=(_document_=document_id) drop=in_category_rule);
    by document_id name;
run;
 
/* delete intermediate data sets */

proc sql noprint;
    drop table _tmp_doc_cat_ds;
    drop table _tmp_transposed_ds;
    drop table _tmp_missing_ds;
    drop table _input_doc_ds;
quit;

/************************************************/
/* Run Text Analytics Sentiment Model in Hadoop */
/************************************************/

options set=SAS_HADOOP_JAR_PATH="/opt/sasinside/hadoop/lib";
options set=SAS_HADOOP_CONFIG_PATH="/opt/sasinside/hadoop/conf";

libname dblib hadoop server = "nodename.demo.xyz.com" user=your.name password=xyz123;

* INPUT DATASET;
%let input_ds = dblib.mydata;

* COLUMN TO PROCESS IN THE INPUT DATASET;
%let document_column = text; 

* OUTPUT DATASET;
%let output_ds = dblib.mydata_out; 

* DELETE OUTPUT BEFORE START, (OPTIONAL);
proc delete data=&output_ds; run;

proc ds2 ds2accel=yes;
    require package tksent; run;
    require package tktxtanio; run;
    require package tkling; run;
    
        thread workerth / overwrite=YES; 

        dcl package tksent sent();
        dcl package tktxtanio txtanio();
        dcl package tkling ling();

        dcl binary(8) _apply_settings;
        dcl binary(8) _sam;
        dcl binary(8) _trans;
        dcl binary(8) _document;
        dcl char(300) _binfile;
        dcl double    _status;
        dcl char(10)  _sentiment_;
        dcl double    _sentiment_probability_;

        retain _apply_settings;
        retain _sam;
        retain _trans;

        /* only runs once when starting */
        method init();
            _status = ling.set_language('english');
            if _status NE 0 then put 'ERROR: set_language fails';

            _status = ling.set_default_language_data();
            if _status NE 0 then put 'ERROR: set_default_language_data fails';

            _apply_settings = sent.new_apply_settings();

            _status = sent.set_threshold(_apply_settings, 0.5);
            if _status NE 0 then put 'ERROR: set_threshold fails';

            _status = sent.set_solution(_apply_settings, 0.0);
            if _status NE 0 then put 'ERROR: set_solution fails';

            _status = sent.set_ratio(_apply_settings, 1.0);
            if _status NE 0 then put 'ERROR: set_ratio fails';

            _sam = txtanio.new_on_content_server('/user_space/en-base.sam');
       
            _status = sent.set_model(_apply_settings, _sam);
            if _status NE 0 then put 'ERROR: set_model fails';

            _trans = sent.new_transaction();

            _status = sent.initialize_sentiment(_apply_settings, _trans);
            if _status NE 0 then put 'ERROR: initialize_sentiment fails';
        end;

        /* run per row of input */
        method run();
            set &input_ds(keep=(&document_column docid));

            _document = txtanio.new_document_from_string(&document_column);
 
            /* set the document on the transaction so we're ready to process */
            _status = sent.set_document(_trans, _document);
            if _status NE 0 then put 'ERROR: set_document fails on document:' docid;

            /* apply the binary to the document */
            _status = sent.apply_sentiment(_apply_settings, _trans);
            if _status NE 0 then put 'ERROR: apply_sentiment fails on document:' docid 'the observation variable is probably empty.';

            _sentiment_ = sent.get_sentiment(_trans);
            _sentiment_probability_ = sent.get_probability(_trans);

            output;

            /* clean up resources */
            sent.clean_sentiment(_trans);
            txtanio.free_object(_document);
        end;

        /* only runs once at end */
        method term();
            /* clean up resources */
            sent.free_transaction(_trans);
            sent.free_apply_settings(_apply_settings);
            txtanio.free_object(_sam);
        end;
    endthread;
run;

data &output_ds(
         keep=(
           docid
           _sentiment_
           _sentiment_probability_
         )
         overwrite=yes
       );
dcl THREAD workerth THRD;
method run();
set from THRD;
end;
enddata;
run;
quit;
